"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.simplify = void 0;
const single_1 = require("./simplifications/single");
const const_enum_1 = require("./simplifications/const-enum");
const intersect_const_enum_1 = require("./simplifications/intersect-const-enum");
const error_1 = require("./error");
const annotation_1 = require("./annotation");
const util_1 = require("./util");
function simplify(node) {
    if (Array.isArray(node))
        return node.map(node => simplify(node));
    if (util_1.isNodeDocument(node))
        return {
            ...node,
            types: simplify(node.types),
        };
    const wrapName = (newNode) => util_1.copyName(node, newNode);
    if (node.type !== 'and' && node.type !== 'or')
        return wrapName(single_1.simplifySingle(node));
    else if (node.type === 'and') {
        const and = simplifyIntersection([].concat(...node.and.map(node => {
            const simplifiedNode = simplify(node);
            return simplifiedNode.and
                ? simplifiedNode.and
                : [simplifiedNode];
        })));
        if (and.length === 1)
            return wrapName(and[0]);
        return wrapName({ type: 'and', and });
    }
    else if (node.type === 'or') {
        const or = simplifyUnion([].concat(...node.or.map(node => {
            const simplifiedNode = simplify(node);
            return simplifiedNode.or
                ? simplifiedNode.or
                : [simplifiedNode];
        })));
        if (or.length === 1)
            return wrapName(or[0]);
        return wrapName({ type: 'or', or });
    }
    else {
        throw new error_1.MalformedTypeError("Invalid node", node);
    }
}
exports.simplify = simplify;
// Combine types/nodes where one is more generic than some other, or where
// they can be combined to fewer nodes.
function simplifyUnion(nodes) {
    const typeMap = util_1.splitTypes(nodes);
    if (typeMap.any.length > 0) {
        const enums = const_enum_1.mergeConstEnumUnion(typeMap.any);
        if (enums.length === 0)
            // If any type in a set of types is an "any" type, without const
            // or enum, the whole union is "any".
            return [{ type: 'any', ...annotation_1.mergeAnnotations(typeMap.any) }];
    }
    if (typeMap.boolean.length === 1)
        typeMap.boolean = [
            single_1.simplifySingle(typeMap.boolean[0])
        ];
    else if (typeMap.boolean.length > 1) {
        const bools = const_enum_1.mergeConstEnumUnion(typeMap.boolean);
        if (bools.length === 0)
            // This enum/const can be removed in favor of generic boolean
            typeMap.boolean = [{ type: 'boolean' }];
        else
            typeMap.boolean = [
                single_1.simplifySingle({ type: 'boolean', enum: bools })
            ];
    }
    if (typeMap.or.length > 0)
        typeMap.or = typeMap.or.filter(({ or }) => or.length > 0);
    if (typeMap.and.length > 0)
        typeMap.and = typeMap.and.filter(({ and }) => and.length > 0);
    return [].concat(...Object.values(typeMap));
}
// Combine types/nodes and exclude types, const and enum where other are
// narrower/stricter.
function simplifyIntersection(nodes) {
    const typeMap = util_1.splitTypes(nodes);
    if (typeMap.any.length > 0) {
        if (typeMap.and.length === 0 &&
            typeMap.or.length === 0 &&
            typeMap.ref.length === 0 &&
            typeMap.null.length === 0 &&
            typeMap.string.length === 0 &&
            typeMap.number.length === 0 &&
            typeMap.integer.length === 0 &&
            typeMap.boolean.length === 0 &&
            typeMap.object.length === 0 &&
            typeMap.array.length === 0 &&
            typeMap.tuple.length === 0)
            return [{ type: 'any', ...annotation_1.mergeAnnotations(typeMap.any) }];
        else
            // A more precise type will supercede this
            typeMap.any = [];
    }
    const cast = (nodes) => nodes;
    if (typeMap.boolean.length > 1)
        typeMap.boolean = [intersect_const_enum_1.intersectConstEnum([
                ...typeMap.boolean,
                ...cast(typeMap.any),
            ])];
    if (typeMap.string.length > 1)
        typeMap.string = [intersect_const_enum_1.intersectConstEnum([
                ...typeMap.string,
                ...cast(typeMap.any),
            ])];
    if (typeMap.number.length > 0 && typeMap.integer.length > 0) {
        typeMap.number = [intersect_const_enum_1.intersectConstEnum([
                ...typeMap.number,
                ...cast(typeMap.integer),
                ...cast(typeMap.any),
            ])];
        typeMap.integer = [];
    }
    else if (typeMap.number.length > 1)
        typeMap.number = [intersect_const_enum_1.intersectConstEnum([
                ...typeMap.number,
                ...cast(typeMap.any),
            ])];
    else if (typeMap.integer.length > 1)
        typeMap.integer = [intersect_const_enum_1.intersectConstEnum([
                ...typeMap.integer,
                ...cast(typeMap.any),
            ])];
    if (typeMap.or.length > 0)
        typeMap.or = typeMap.or.filter(({ or }) => or.length > 0);
    if (typeMap.and.length > 0)
        typeMap.and = typeMap.and.filter(({ and }) => and.length > 0);
    return [].concat(...Object.values(typeMap));
}
