export interface LineColumn {
    line: number;
    column: number;
    offset: number;
}
export interface LocationWithLineColumn {
    start?: LineColumn;
    end?: LineColumn;
}
export interface LocationOffset {
    start?: number;
    end?: number;
}
export declare type Location = LocationWithLineColumn | LocationOffset;
export interface CoreTypeAnnotations {
    name?: string;
    title?: string;
    description?: string;
    examples?: string | Array<string>;
    default?: string;
    see?: string | Array<string>;
    comment?: string;
    loc?: LocationOffset | LocationWithLineColumn;
}
export interface AndType extends CoreTypeAnnotations {
    type: 'and';
    and: NodeType[];
}
export interface OrType extends CoreTypeAnnotations {
    type: 'or';
    or: NodeType[];
}
export interface TypeMap {
    and: void;
    or: void;
    ref: string;
    any: unknown;
    null: null;
    string: string;
    number: number;
    integer: number;
    boolean: boolean;
    object: object;
    array: Array<unknown>;
    tuple: Array<unknown>;
}
export declare type Types = keyof TypeMap;
export interface Const<T> {
    const?: T;
}
export interface Enum<T> {
    enum?: Array<T>;
}
export declare type GenericTypeInfo<T> = Const<T> & Enum<T>;
export interface NodePrimitiveCoreType<Type extends 'any' | 'null' | 'string' | 'number' | 'integer' | 'boolean'> {
    type: Type;
}
export declare type NodePrimitiveType<Type extends 'any' | 'null' | 'string' | 'number' | 'integer' | 'boolean'> = NodePrimitiveCoreType<Type> & GenericTypeInfo<TypeMap[Type]>;
export declare type AnyType = NodePrimitiveType<'any'> & CoreTypeAnnotations;
export declare type NullType = NodePrimitiveCoreType<'null'> & CoreTypeAnnotations;
export declare type StringType = NodePrimitiveType<'string'> & CoreTypeAnnotations;
export declare type NumberType = NodePrimitiveType<'number'> & CoreTypeAnnotations;
export declare type IntegerType = NodePrimitiveType<'integer'> & CoreTypeAnnotations;
export declare type BooleanType = NodePrimitiveType<'boolean'> & CoreTypeAnnotations;
export declare type PrimitiveType = NullType | StringType | NumberType | IntegerType | BooleanType;
export interface NodeRefCoreType {
    type: 'ref';
    ref: string;
}
export declare type RefType = NodeRefCoreType & GenericTypeInfo<unknown> & CoreTypeAnnotations;
export interface ObjectProperty {
    required: boolean;
    node: NodeType;
}
export interface NodeObjectCoreType {
    type: 'object';
    properties: {
        [name: string]: ObjectProperty;
    };
    additionalProperties: boolean | NodeType;
}
export declare type ObjectType = NodeObjectCoreType & GenericTypeInfo<object> & CoreTypeAnnotations;
export interface NodeArrayCoreType {
    type: 'array';
    elementType: NodeType;
}
export declare type ArrayType<T = unknown> = NodeArrayCoreType & GenericTypeInfo<Array<T>> & CoreTypeAnnotations;
export interface NodeTupleCoreType {
    type: 'tuple';
    elementTypes: Array<NodeType>;
    minItems: number;
    additionalItems: boolean | NodeType;
}
export declare type TupleType<T extends unknown[] = unknown[]> = NodeTupleCoreType & GenericTypeInfo<T> & CoreTypeAnnotations;
export interface NodeTypeMap {
    and: AndType;
    or: OrType;
    ref: RefType;
    any: AnyType;
    null: NullType;
    string: StringType;
    number: NumberType;
    integer: IntegerType;
    boolean: BooleanType;
    object: ObjectType;
    array: ArrayType;
    tuple: TupleType;
}
export declare type NodeType = NodeTypeMap[keyof NodeTypeMap];
export declare type NamedType<T extends NodeType = NodeType> = T & {
    name: string;
};
export declare type NodeWithConstEnum = AnyType | StringType | NumberType | IntegerType | BooleanType | ObjectType | ArrayType | TupleType | RefType;
export declare type NodePath = Array<string | number>;
export interface NodeDocument<Version extends number = 1, T extends NodeType = NodeType> {
    version: Version;
    types: Array<NamedType<T>>;
}
export interface ConversionResult<T = string> {
    data: T;
    convertedTypes: Array<string>;
    notConvertedTypes: Array<string>;
}
